"""
   lookupy
   ~~~~~~~

   Lookupy is a Python library that provides a Django QuerySet like
   interface to query (select and filter) data (list of dicts)

"""

from .lookupy import Collection, Q

__all__ = ["Collection", "Q"]

